"""
Utilities for unit tests
"""

import pytest
import torch

CUDA_AVAILABLE = torch.cuda.is_available()


def needs_cuda(method):
    """
    Decorator to xfail methods that require CUDA when we don't have it.
    """
    if CUDA_AVAILABLE:
        return method
    else:
        @pytest.mark.xfail(reason="CUDA not available")
        def wrapped(obj, *args, **kwargs):
            method(obj, *args, **kwargs)
        
        return wrapped


# File: settings.py
# File Created: Thursday, 11th July 2019 12:13:30 am
# Author: Steven Atkinson (steven@atkinson.mn)

from torch.distributions.transforms import ExpTransform

DefaultPositiveTransform = ExpTransform



# File: __init__.py
# Author: Yinhao Zhu (yzhu10@nd.edu)

from . import model
from . import functions
from . import mean_functions
from . import kernels
from . import ekernels
from . import likelihoods
from . import param
from . import settings
from . import util

from . import models



#
# Yinhao Zhu, May 01, 2017
#
"""
Sparse GP regression, including variational GP and others.
"""

from __future__ import absolute_import

import torch
import numpy as np
from torch.utils.data import TensorDataset, DataLoader
from torch.distributions.transforms import LowerCholeskyTransform

from ..model import Param
from ..functions import cholesky, trtrs
from ..mean_functions import Zero
from ..likelihoods import Gaussian
from ..util import TensorType, torch_dtype, as_tensor, kmeans_centers
from .gpr import GPR
from .base import GPModel


class _InducingPointsGP(GPModel):
    """
    Parent class for GPs with inducing points
    """

    def __init__(
        self,
        x,
        y,
        kernel,
        num_inducing_points=None,
        inducing_points=None,
        mean_function=None,
        likelihood=None,
    ):
        """
        Assume Gaussian likelihood

        Args:
            observations (np.ndarray): Y, n x p
            input (np.ndarray): X, n x q
            kernel (gptorch.Kernel):
            inducing_points (np.ndarray, optional): Z, m x q
            num_inducing (int), optional): number of inducing inputs

        Input, observations, and kernel must be specified, if both
        ``inducing_points`` and ``num_inducing`` are not set, 1/10 th of total
        points (up to 100) will be draw randomly from input as the inducing 
        points.
        """

        super().__init__(x, y, kernel, likelihood, mean_function)

        if inducing_points is None:
            if num_inducing_points is None:
                num_inducing_points = np.clip(x.shape[0] // 10, 1, 100)
            inducing_points = kmeans_centers(x, num_inducing_points, 
                perturb_if_fail=True)
            # indices = np.random.permutation(len(x))[:num_inducing_points]
            # inducing_points = TensorType(x[indices])

        # Z stands for inducing input points as standard in the literature
        self.Z = Param(as_tensor(inducing_points))

    @property
    def num_inducing(self) -> int:
        """
        Number of inducing points
        """
        return self.Z.shape[0]


class FITC(_InducingPointsGP):
    """
    Fully Independent Training Conditional approximation for GP

    References:
        Snelson, Edward, and Zoubin Ghahramani. "Sparse Gaussian processes
         using pseudo-inputs." Advances in neural information processing
         systems 18 (2006): 1257.
        Quinonero-Candela, Joaquin, and Carl Edward Rasmussen. "A unifying
        view of sparse approximate Gaussian process regression." Journal of
        Machine Learning Research 6.Dec (2005): 1939-1959.
    """

    # TODO: add FITC for sparse GP regression
    pass


class VFE(_InducingPointsGP):
    """
    Variational Free Energy approximation for GP

    Reference:
        Titsias, Michalis K. "Variational Learning of Inducing Variables
        in Sparse Gaussian Processes." AISTATS. Vol. 5. 2009.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(
            self.mean_function, Zero
        ), "Mean functions not implemented for VFE yet."

    def log_likelihood(self, x=None, y=None):
        """
        Computes the variational lower bound of the true log marginal likelihood
        Eqn (9) in Titsias, Michalis K. "Variational Learning of Inducing Variables
        in Sparse Gaussian Processes." AISTATS. Vol. 5. 2009.
        """

        x = x if x is not None else self.X
        y = y if y is not None else self.Y
        if not x.shape[0] == y.shape[0]:
            raise ValueError("X and Y must have same # data.")

        num_inducing = self.num_inducing
        num_data = x.shape[0]
        d_out = self.output_dimension
        # TODO: add mean_functions
        # err = self.Y - self.mean_function(x)
        err = self.Y
        Kff_diag = self.kernel.Kdiag(x)
        Kuf = self.kernel.K(self.Z, x)
        # add jitter
        Kuu = self.kernel.K(self.Z)
        L = cholesky(Kuu)

        A = trtrs(Kuf, L)
        AAT = A @ A.t() / self.likelihood.variance.transform().expand_as(Kuu)
        B = AAT + torch.eye(num_inducing, dtype=torch_dtype).to(AAT.device)
        LB = cholesky(B)
        # divide variance at the end
        c = trtrs(A @ err, LB) / self.likelihood.variance.transform()

        # Evidence lower bound
        elbo = TensorType([-0.5 * d_out * num_data * np.log(2 * np.pi)]).to(c.device)
        elbo -= d_out * LB.diag().log().sum()
        elbo -= (
            0.5 * d_out * num_data * self.likelihood.variance.transform().log()
        )
        elbo -= (
            0.5
            * (err.pow(2).sum() + d_out * Kff_diag.sum())
            / self.likelihood.variance.transform()
        )
        elbo += 0.5 * c.pow(2).sum()
        elbo += 0.5 * d_out * AAT.diag().sum()

        return elbo[0]

    def _predict(self, x_new: TensorType, diag=True, x=None):
        """
        Compute posterior p(f*|y), integrating out induced outputs' posterior.

        :return: (mean, var/cov)
        """

        x = x if x is not None else self.X

        z = self.Z
        z.requires_grad_(False)

        num_inducing = z.size(0)

        # err = self.Y - self.mean_function(x)
        err = self.Y
        Kuf = self.kernel.K(z, x)
        # add jitter
        Kuu = self.kernel.K(z)
        Kus = self.kernel.K(z, x_new)
        L = cholesky(Kuu)
        A = trtrs(Kuf, L)
        AAT = A @ A.t() / self.likelihood.variance.transform().expand_as(Kuu)
        B = AAT + torch.eye(num_inducing, dtype=torch_dtype).to(AAT.device)
        LB = cholesky(B)
        # divide variance at the end
        c = trtrs(A @ err, LB) / self.likelihood.variance.transform()
        tmp1 = trtrs(Kus, L)
        tmp2 = trtrs(tmp1, LB)
        mean = tmp2.t() @ c

        if diag:
            var = (
                self.kernel.Kdiag(x_new)
                - tmp1.pow(2).sum(0).squeeze()
                + tmp2.pow(2).sum(0).squeeze()
            )[:, None].expand_as(mean)
        else:
            var = self.kernel.K(x_new) + tmp2.t() @ tmp2 - tmp1.t() @ tmp1

        return mean, var


def minibatch(loss_func):
    """
    Decorator to use minibatching for a loss function (e.g. SVGP)
    """

    def wrapped(obj, x=None, y=None):
        if x is not None:
            assert y is not None
        else:
            # Get from model:
            if obj.batch_size is not None:
                i = np.random.permutation(obj.num_data)[: obj.batch_size]
                x, y = obj.X[i, :], obj.Y[i, :]
            else:
                x, y = obj.X, obj.Y

        return loss_func(obj, x, y)

    return wrapped


class SVGP(_InducingPointsGP):
    """
    Sparse variational Gaussian process.

    James Hensman, Nicolo Fusi, and Neil D. Lawrence,
    "Gaussian processes for Big Data" (2013)

    James Hensman, Alexander Matthews, and Zoubin Ghahramani, 
    "Scalable variational Gaussian process classification", JMLR (2015).
    """

    def __init__(
        self,
        y,
        x,
        kernel,
        num_inducing_points=None,
        inducing_points=None,
        mean_function=None,
        likelihood=Gaussian(),
        batch_size=None,
    ):
        """
        :param batch_size: How many points to process in a minibatch of 
            training.  If None, no minibatches are used.
        """
        super().__init__(
            y,
            x,
            kernel,
            num_inducing_points=num_inducing_points,
            inducing_points=inducing_points,
            mean_function=mean_function,
            likelihood=likelihood,
        )
        # assert batch_size is None, "Minibatching not supported yet."
        self.batch_size = batch_size

        # Parameters for the Gaussian variational posterior over the induced
        # outputs.
        # Note: induced_output_mean does NOT include the contribution due to the
        # mean function.
        self.induced_output_mean, self.induced_output_chol_cov = self._init_posterior()

    @minibatch
    def log_likelihood(self, x, y):
        """
        Variational bound.
        """

        if not x.shape[0] == y.shape[0]:
            raise ValueError("X and Y must have same # data.")

        chol_kuu = cholesky(self.kernel.K(self.Z))

        # Marginal posterior q(f)'s mean & variance
        f_mean, f_var = self._predict(x, diag=True, chol_kuu=chol_kuu)
        marginal_log_likelihood = torch.stack(
            [
                self.likelihood.propagate_log(
                    torch.distributions.Normal(loc_i, torch.sqrt(v_i)), yi
                )
                for loc_i, v_i, yi in zip(f_mean.t(), f_var.t(), y.t())
            ]
        ).sum()
        # Account for size of minibatch relative to the total dataset size:
        marginal_log_likelihood *= self.num_data / x.shape[0]

        mu_xu = self.mean_function(self.Z)  # Prior mean
        qu_mean = self.induced_output_mean + mu_xu
        qu_lc = self.induced_output_chol_cov.transform()
        # Each output dimension has its own Multivariate normal (different
        # means, shared covariance); the joint distribution is the product
        # across output dimensions.
        qus = [
            torch.distributions.MultivariateNormal(qu_i, scale_tril=qu_lc)
            for qu_i in qu_mean.t()
        ]
        # Each dimension has its own prior as well due to the mean function
        # Being potentially different for each output dimension.
        pus = [
            torch.distributions.MultivariateNormal(mi, scale_tril=chol_kuu)
            for mi in mu_xu.t()
        ]

        kl = torch.stack(
            [torch.distributions.kl_divergence(qu, pu) for qu, pu in zip(qus, pus)]
        ).sum()

        return marginal_log_likelihood - kl

    def _init_posterior(self):
        """
        Get an initial guess at the variational posterior over the induced 
        outputs.

        Just build a GP out of a few data and use its posterior.
        This could be far worse than expected if the likelihood is non-Gaussian,
        but we don't need this to be great--just good enough to get started.
        """

        i = np.random.permutation(self.num_data)[0 : min(self.num_data, 100)]
        x, y = self.X[i].data.numpy(), self.Y[i].data.numpy()
        # Likelihood needs to be Gaussian for exact inference in GPR
        likelihood = (
            self.likelihood
            if isinstance(self.likelihood, Gaussian)
            else Gaussian(variance=0.01 * y.var())
        )
        model = GPR(
            x, y, self.kernel, mean_function=self.mean_function, likelihood=likelihood
        )
        mean, cov = model.predict_f(self.Z, diag=False)
        mean -= self.mean_function(self.Z)
        chol_cov = cholesky(cov)

        return Param(mean), Param(chol_cov, transform=LowerCholeskyTransform())

    def _predict(self, x_new: TensorType, diag=True, chol_kuu=None, **kwargs):
        """
        SVGP Prediction uses inducing points as sufficient statistics for the 
        posterior.

        Could implement Marginalization of Gaussians (cf. PRML p. 93), but
        something specific to (positive-definite) kernel matrices should 
        perform better.

        Shapes of outputs are:
        diag: both are [N x dy]
        not diag: mean is [N x dy], cov is [N x N]

        :param x_new: inputs to predict on.
        :param diag: if True, return variance of prediction; False=full cov
        :param chol_kuu: The Cholesky of the kernel matrix for the inducing 
            inputs (to enable reuse when computing the training loss)
    
        :return: (torch.Tensor, torch.Tensor) mean & [co]variance
        """

        chol_kuu = cholesky(self.kernel.K(self.Z)) if chol_kuu is None else chol_kuu
        kuf = self.kernel.K(self.Z, x_new)
        alpha = trtrs(kuf, chol_kuu).t()
        # beta @ beta.t() = inv(L) @ S @ inv(L'), S=post cov of induced outs
        beta = trtrs(self.induced_output_chol_cov.transform(), chol_kuu)
        mu_x = self.mean_function(x_new)

        # Remember: induced_output_mean doesn't include mean function, so no
        # need to subtract it.
        f_mean = alpha @ trtrs(self.induced_output_mean, chol_kuu) + mu_x

        # gamma @ gamma.t() = Kfu @ inv(Kuu) @ S @ inv(Kuu) @ Kuf
        gamma = alpha @ beta

        if diag:
            f_cov = (
                self.kernel.Kdiag(x_new)
                - torch.sum(alpha ** 2, dim=1)
                + torch.sum(gamma ** 2, dim=1)
            )[:, None].expand_as(f_mean)
        else:
            f_cov = self.kernel.K(x_new) - alpha @ alpha.t() + gamma @ gamma.t()

        return f_mean, f_cov



# File: test_gpr.py
# File Created: Saturday, 13th July 2019 3:25:43 pm
# Author: Steven Atkinson (steven@atkinson.mn)

import os
import sys

import pytest
import numpy as np
import torch

from gptorch.models import GPR
from gptorch.kernels import Rbf
from gptorch.util import torch_dtype, TensorType

base_path = os.path.join(os.path.dirname(__file__), "..", "..")
if not base_path in sys.path:
    sys.path.append(base_path)

from test.util import needs_cuda


class TestGPR(object):
    def test_init(self):
        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(x.shape[1], ARD=True)

        # init w/ numpy
        GPR(x, y, kern)
        # init w/ PyTorch tensors:
        GPR(TensorType(y), TensorType(x), kern)
        # init w/ a mean function:
        GPR(x, y, kern, mean_function=torch.nn.Linear(dx, dy))

    def test_loss(self):
        model, x, y = self._get_model()
        n = x.shape[0]

        loss = model.loss()
        assert isinstance(loss, torch.Tensor)
        assert loss.ndimension() == 1  # TODO change this...

        # Test ability to specify x and y
        loss_xy = model.loss(x=TensorType(x), y=TensorType(y))
        assert isinstance(loss_xy, torch.Tensor)
        assert loss_xy.item() == loss.item()

        with pytest.raises(ValueError):
            # Size mismatch
            model.loss(x=TensorType(x[:n // 2]))

    @needs_cuda
    def test_compute_loss_cuda(self):
        model = self._get_model()[0]
        model.cuda()

        loss = model.loss()
        assert loss.is_cuda

    def test_predict(self):
        n, n_test, dx, dy = 5, 7, 3, 2
        x, y = torch.randn(n, dx), torch.randn(n, dy)
        kern = Rbf(x.shape[1], ARD=True)
        model = GPR(x, y, kern)

        x_test = torch.randn(n_test, dx)
        mu_var, var = model._predict(x_test)
        assert all([e == a for e, a in zip(mu_var.shape, (n_test, dy))])
        assert all([e == a for e, a in zip(var.shape, (n_test, dy))])

        mu_cov, cov = model._predict(x_test, diag=False)
        assert all([e == a for e, a in zip(mu_cov.shape, (n_test, dy))])
        assert all([e == a for e, a in zip(cov.shape, (n_test, n_test))])

    @needs_cuda
    def test_predict_cuda(self):
        n, n_test, dx, dy = 5, 7, 3, 2
        x, y = torch.randn(n, dx), torch.randn(n, dy)
        kern = Rbf(x.shape[1], ARD=True)
        model = GPR(x, y, kern)
        model.cuda()

        x_test = torch.randn(n_test, dx, dtype=torch_dtype).cuda()
        for t in model._predict(x_test):  # mean, variance
            assert t.is_cuda

    @staticmethod
    def _get_model():
        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(x.shape[1], ARD=True)

        return GPR(x, y, kern), x, y



# File: common.py
# File Created: Sunday, 1st September 2019 9:57:32 am
# Author: Steven Atkinson (steven@atkinson.mn)

import torch
import numpy as np
import pytest

from gptorch.models.base import GPModel
from gptorch.util import torch_dtype

torch.set_default_dtype(torch_dtype)


def gaussian_predictions(model: GPModel, x_test: torch.Tensor, 
        expected_mu: np.ndarray, expected_s: np.ndarray):
    """
    Every GP model with a gaussian likelihood needs the same set of tests run on
    its ._predict() method.
    """

    # Predictions without full covariance
    mu_diag, s_diag = model._predict(x_test, diag=True)

    assert isinstance(mu_diag, torch.Tensor)
    assert isinstance(s_diag, torch.Tensor)

    assert mu_diag.shape[0] == x_test.shape[0]
    assert mu_diag.shape[1] == model.Y.shape[1]
    assert all([ss == ms for ss, ms in zip(mu_diag.shape, s_diag.shape)])

    assert all([a == pytest.approx(e) for a, e in zip(
        mu_diag.detach().numpy().flatten(), expected_mu.flatten())])
    assert all([a == pytest.approx(e) for a, e in zip(
        s_diag.detach().numpy().flatten(), expected_s.diagonal().flatten())])

    # Predictions with full covariance
    mu_full, s_full = model._predict(x_test, diag=False)

    assert isinstance(mu_full, torch.Tensor)
    assert isinstance(s_full, torch.Tensor)

    assert mu_full.shape[0] == x_test.shape[0]
    assert mu_full.shape[1] == model.Y.shape[1]
    assert all([ss == x_test.shape[0] for ss in s_full.shape])
    
    assert all([a == pytest.approx(e) for a, e in zip(
        mu_full.detach().numpy().flatten(), expected_mu.flatten())])
    assert all([a == pytest.approx(e) for a, e in zip(
        s_full.detach().numpy().flatten(), expected_s.flatten())])



# File: test_base.py
# File Created: Saturday, 13th July 2019 1:54:18 pm
# Author: Steven Atkinson (steven@atkinson.mn)

import os
import sys

import pytest
import torch
import numpy as np

from gptorch.models.base import GPModel
from gptorch.mean_functions import Zero
from gptorch.util import TensorType
from gptorch.kernels import Rbf
from gptorch.models import GPR

base_path = os.path.join(os.path.dirname(__file__), "..", "..")
if not base_path in sys.path:
    sys.path.append(base_path)

from test.util import needs_cuda


class TestGPModel(object):
    """
    Tests for the GPModel class
    """

    @needs_cuda
    def test_cuda(self):
        gp = self._get_model()

        gp.cuda()
        # Ensure that the data made it too
        assert gp.X.is_cuda
        assert gp.Y.is_cuda

    @needs_cuda
    def test_cpu(self):
        gp = self._get_model()

        gp.cuda()
        gp.cpu()

        # Ensure that the data made it too
        assert not gp.X.is_cuda
        assert not gp.Y.is_cuda

    def test_optimize(self):
        gp = self._get_model()
        gp.optimize(max_iter=2)
        gp.optimize(method="L-BFGS-B", max_iter=2)

    def test_predict_f(self):
        self._predict_fy("predict_f")

    @needs_cuda
    def test_predict_f_cuda(self):
        self._predict_fy_cuda("predict_f")

    def test_predict_y(self):
        self._predict_fy("predict_y")

    @needs_cuda
    def test_predict_y_cuda(self):
        self._predict_fy_cuda("predict_y")

    def test_predict_f_samples(self):
        self._predict_fy_samples("predict_f_samples")

    @needs_cuda
    def test_predict_f_samples_cuda(self):
        self._predict_fy_samples_cuda("predict_f_samples")

    def test_predict_y_samples(self):
        self._predict_fy_samples("predict_y_samples")

    @needs_cuda
    def test_predict_y_samples_cuda(self):
        self._predict_fy_samples_cuda("predict_y_samples")

    def _predict_fy(self, attr):
        """
        attr='predict_f' or 'predict_y'
        """

        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(dx, ARD=True)
        gp = GPR(x, y, kern)

        n_test = 5
        x_test = np.random.randn(n_test, dx)
        f = getattr(gp, attr)
        mu, v = f(x_test)
        for result in [mu, v]:
            assert isinstance(result, np.ndarray)
            assert result.ndim == 2  # [n_test x dy]
            assert result.shape == (n_test, dy)

        x_test_torch = TensorType(x_test)
        mu_torch, v_torch = f(x_test_torch)
        for result in [mu_torch, v_torch]:
            assert isinstance(result, TensorType)
            assert result.ndimension() == 2  # [n_test x dy]
            assert result.shape == (n_test, dy)

    def _predict_fy_cuda(self, attr):
        """
        attr='predict_f' or 'predict_y'
        """

        gp = self._get_model()
        f = getattr(gp, attr)
        x_test = np.random.randn(5, gp.input_dimension)
        x_test_torch = TensorType(x_test)

        # Test that CUDA works in all cases:
        gp.cuda()
        # Numpy input:
        cuda_np = f(x_test)
        for result in cuda_np:
            assert isinstance(result, np.ndarray)
        # PyTorch (cpu) input
        cuda_torch = f(x_test_torch)
        for result in cuda_torch:
            assert result.device == x_test_torch.device
        # PyTorch (GPU) input
        cuda_gpu = f(x_test_torch.to("cuda"))
        for result in cuda_gpu:
            assert result.is_cuda

    def _predict_fy_samples(self, attr):
        """
        attr="predict_f_samples" or "predict_y_samples"
        """

        # TODO mock a GPModel?  Using GPR for the moment.
        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(dx, ARD=True)
        gp = GPR(x, y, kern)
        f = getattr(gp, attr)

        n_test = 5
        x_test = np.random.randn(n_test, dx)
        samples = f(x_test)
        assert isinstance(samples, np.ndarray)
        assert samples.ndim == 3  # [sample x n_test x dy]
        assert samples.shape == (1, n_test, dy)

        n_samples = 3
        samples_2 = f(x_test, n_samples=n_samples)
        assert isinstance(samples_2, np.ndarray)
        assert samples_2.ndim == 3  # [sample x n_test x dy]
        assert samples_2.shape == (n_samples, n_test, dy)

        x_test_torch = TensorType(x_test)
        samples_torch = f(x_test_torch)
        assert isinstance(samples_torch, TensorType)
        assert samples_torch.ndimension() == 3  # [1 x n_test x dy]
        assert samples_torch.shape == (1, n_test, dy)

    def _predict_fy_samples_cuda(self, attr):
        
        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(dx, ARD=True)
        gp = GPR(x, y, kern)
        f = getattr(gp, attr)

        n_test = 5
        x_test = np.random.randn(n_test, dx)
        x_test_torch = TensorType(x_test)

        gp.cuda()
        # Numpy input:
        samples_cuda_np = f(x_test)
        assert isinstance(samples_cuda_np, np.ndarray)
        # PyTorch (cpu) input
        samples_cuda_torch = f(x_test_torch)
        assert samples_cuda_torch.device == x_test_torch.device
        # PyTorch (GPU) input
        samples_cuda_gpu = f(x_test_torch.to("cuda"))
        assert samples_cuda_gpu.is_cuda

    @staticmethod
    def _get_model():
        n, dx, dy = 5, 3, 2
        x, y = np.random.randn(n, dx), np.random.randn(n, dy)
        kern = Rbf(dx, ARD=True)
        return GPR(x, y, kern)



# File: mean_functions.py
# File Created: Saturday, 13th July 2019 3:42:16 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
mean_functions.py: A few extra mean functions that you don't usually see as
PyTorch modules
"""

import torch

from .util import torch_dtype

torch.set_default_dtype(torch_dtype)


class Constant(torch.nn.Module):
    """
    Constant mean function
    """

    def __init__(self, dy: int, val: torch.Tensor = None):
        super().__init__()
        if val is not None:
            if not val.shape[0] == dy:
                raise ValueError("Provided val doesn't match output dimension")
            val = val.clone()
        else:
            val = torch.zeros(dy)

        self._dy = dy
        self.val = torch.nn.Parameter(val)

    def forward(self, x):
        output = torch.zeros(x.shape[0], self._dy)
        if self._is_cuda():
            output = output.cuda()
        return output + self.val

    def _is_cuda(self):
        return self.val.is_cuda


class Zero(Constant):
    """
    Zero mean function (default for GPs).
    """

    def __init__(self, dy: int):
        super().__init__(dy)
        self.val.requires_grad_(False)



# File: base.py
# File Created: Saturday, 13th July 2019 1:02:55 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
base.py: The core GP model (GP+likelihood)
"""

import numpy as np
import torch
from time import time
from scipy.optimize import minimize

from .. import likelihoods
from ..functions import cholesky
from ..mean_functions import Zero
from ..model import Model
from ..util import TensorType, as_tensor, torch_dtype


def input_as_tensor(predict_func):
    """
    Decorator for prediction funtions to ensure that inputs are TensorType and
    on the correct device before passing into GPModel._predict() methods.

    :param predict_func: The public predict funciton to be wrapped
    """

    def predict(obj, input_new, *args, **kwargs):
        from_numpy = isinstance(input_new, np.ndarray)
        if from_numpy:
            input_new = TensorType(input_new)
            input_new = input_new.to(obj.Y.device)  # Assume single GPU
        else:
            # Ensure we match device:
            outside_device = input_new.device
            input_new = input_new.to(obj.Y.device)
        out = predict_func(obj, input_new, *args, **kwargs)
        if from_numpy:
            if isinstance(out, torch.Tensor):
                out = out.detach().cpu().numpy()
            elif isinstance(out, tuple):
                out = tuple([o.detach().cpu().numpy() for o in out])
            else:
                raise NotImplementedError("Unhandled output type {}".format(type(out)))
        else:
            if isinstance(out, torch.Tensor):
                out = out.to(outside_device)
            elif isinstance(out, tuple):
                out = tuple([o.to(outside_device) for o in out])
            else:
                raise NotImplementedError("Unhandled output type {}".format(type(out)))
        return out

    return predict


class GPModel(Model):
    """
    The base class for GP models
    """

    def __init__(self, x, y, kernel, likelihood, mean_function, name="gp"):
        """
        For unsupervised case, x is optional ...
        Args:
            y (ndarray): Y, N x q
            x (ndarray): X, N x p
            kernel (gptorch.Kernel):
            likelihood (gptorch.Likelihood):
            mean_function (gptorch.MeanFunction):
            name (string): name of this model
        """

        super().__init__()
        self.kernel = kernel
        self.likelihood = likelihood if likelihood is not None else \
            GPModel._init_gaussian_likelihood(y)
        self.mean_function = mean_function if mean_function is not None else \
            Zero(y.shape[1])

        x, y = as_tensor(x), as_tensor(y)
        x.requires_grad_(False)
        y.requires_grad_(False)
        self.X, self.Y = x, y

        self.__class__.__name__ = name

    @property
    def num_data(self):
        return self.Y.shape[0]

    @property
    def input_dimension(self):
        return self.X.shape[1]

    @property
    def output_dimension(self):
        return self.Y.shape[1]

    @staticmethod
    def _init_gaussian_likelihood(y) -> likelihoods.Gaussian:
        """
        A handy heuristic for initializing Gaussian likelihoods for models: make
        the standard deviation roughly 3% of the total output variance.

        :param y: Outputs.  Can be either TensorType or np.ndarray
        """
        return likelihoods.Gaussian(variance=0.001 * y.var())

    def optimize(self, method='Adam', max_iter=2000, verbose=True,
            learning_rate=None):
        """
        Optimizes the model by minimizing the loss (from :method:) w.r.t.
        model parameters.
        Args:
            method (torch.optim.Optimizer, optional): Optimizer in PyTorch
                (maybe add scipy optimizer in the future), default is `Adam`.
            max_iter (int): Max iterations, default 2000.
            verbose (bool, optional): Shows more details on optimization
                process if True.
        Todo:
            Add stochastic optimization, such as mini-batch.
        Returns:
            (np.array, value):
                losses: losses over optimization steps, (max_iter, )
                time: time taken approximately
        """
        parameters = filter(lambda p: p.requires_grad, self.parameters())

        default_learning_rates = {
            "SGD": 0.001,
            "Adam": 0.01,
            "LBFGS": 1.0,
            "Adadelta": 1.0,
            "Adagrad": 0.01,
            "Adamax": 0.002,
            "ASGD": 0.01,
            "RMSprop": 0.01,
            "Rprop": 0.01
        }
        if learning_rate is None and method in default_learning_rates:
            learning_rate = default_learning_rates[method]
        if method == "SGD":
            # GPs seem to benefit from a more aggressive learning rate than the
            # usual "0.001 or so" that NNs like.
            learning_rate = learning_rate if learning_rate is not None else 0.01
            self.optimizer = torch.optim.SGD(parameters, lr=learning_rate, momentum=0.9)
        elif method == "Adam":
            learning_rate = learning_rate if learning_rate is not None else 0.01
            self.optimizer = torch.optim.Adam(parameters, lr=learning_rate)
        elif method == "LBFGS":
            if learning_rate is None:
                learning_rate = 1.0
            self.optimizer = torch.optim.LBFGS(
                parameters,
                lr=learning_rate,
                max_iter=5,
                max_eval=None,
                tolerance_grad=1e-05,
                tolerance_change=1e-09,
                history_size=50,
                line_search_fn=None,
            )
        elif method == "Adadelta":
            self.optimizer = torch.optim.Adadelta(
                parameters, lr=learning_rate, rho=0.9, eps=1e-06, 
                weight_decay=0.00001
            )
        elif method == "Adagrad":
            self.optimizer = torch.optim.Adagrad(
                parameters, lr=learning_rate, lr_decay=0, weight_decay=0
            )
        elif method == "Adamax":
            self.optimizer = torch.optim.Adamax(
                parameters, lr=learning_rate, betas=(0.9, 0.999), eps=1e-08, weight_decay=0
            )
        elif method == "ASGD":
            self.optimizer = torch.optim.ASGD(
                parameters,
                lr=learning_rate,
                lambd=0.0001,
                alpha=0.75,
                t0=1000000.0,
                weight_decay=0,
            )
        elif method == "RMSprop":
            self.optimizer = torch.optim.RMSprop(
                parameters,
                lr=learning_rate,
                alpha=0.99,
                eps=1e-08,
                weight_decay=0.00,
                momentum=0.01,
                centered=False,
            )
        elif method == "Rprop":
            self.optimizer = torch.optim.Rprop(
                parameters, lr=learning_rate, etas=(0.5, 1.2), step_sizes=(1e-06, 50)
            )
        # scipy.optimize.minimize
        # suggest to use L-BFGS-B, BFGS
        elif method in [
            "CG",
            "BFGS",
            "Newton-CG",
            "Nelder-Mead",
            "Powell",
            "L-BFGS-B",
            "TNC",
            "COBYLA",
            "SLSQP",
            "dogleg",
            "trust-ncg",
        ]:
            print("Scipy.optimize.minimize...")
            return self._optimize_scipy(method=method, maxiter=max_iter, disp=verbose)

        else:
            raise ValueError(
                "Optimizer %s is not found. Please choose one of the"
                "following optimizers supported in PyTorch:"
                "Adadelt, Adagrad, Adam, Adamax, ASGD, LBFGS, "
                "RMSprop, Rprop, SGD, LBFGS. Or the optimizers "
                "supported scipy.optimize.minminze: BFGS, L-BFGS-B,"
                "CG, Newton-CG, Nelder-Mead, Powell, TNC, COBYLA,"
                "SLSQP, dogleg, trust-ncg, etc." % method
            )

        losses = np.zeros(max_iter)
        tic = time()

        print("{}: Start optimizing via {}".format(self.__class__.__name__, method))
        if verbose:
            if not method == "LBFGS":
                for idx in range(max_iter):
                    self.optimizer.zero_grad()
                    # forward
                    loss = self.loss()
                    # backward
                    loss.backward()
                    self.optimizer.step()
                    losses[idx] = loss.item()
                    print("Iter: %d\tLoss: %s" % (idx, loss.item()))

            else:
                for idx in range(max_iter):

                    def closure():
                        self.optimizer.zero_grad()
                        loss = self.loss()
                        loss.backward()
                        return loss

                    loss = self.optimizer.step(closure)
                    losses[idx] = loss.item()
                    print("Iter: %d\tLoss: %s" % (idx, loss.item()))
        else:
            if not method == "LBFGS":
                for idx in range(max_iter):
                    self.optimizer.zero_grad()
                    # forward
                    loss = self.loss()
                    # backward
                    loss.backward()
                    self.optimizer.step()
                    losses[idx] = loss.item()
                    if idx % 20 == 0:
                        print("Iter: %d\tLoss: %s" % (idx, loss.item()))
            else:
                for idx in range(max_iter):

                    def closure():
                        self.optimizer.zero_grad()
                        loss = self.loss()
                        loss.backward()
                        return loss

                    loss = self.optimizer.step(closure)
                    if isinstance(loss, float):  # Converged!
                        losses[idx] = loss
                        losses = losses[0 : idx + 1]
                        break
                    else:
                        losses[idx] = loss.item()
                    if idx % 20 == 0:
                        print("Iter: %d\tLoss: %s" % (idx, losses[idx]))
        t = time() - tic
        print("Optimization time taken: %s s" % t)
        print("Optimization method: %s" % str(self.optimizer))
        if len(losses) == max_iter:
            print("Optimization terminated by reaching the maximum iterations")
        else:
            print("Optimization terminated by getting below the tolerant error")

        return losses, t

    def _optimize_scipy(
        self, method="L-BFGS-B", tol=None, callback=None, maxiter=1000, disp=True
    ):
        """
        Wrapper of scipy.optimize.minimize
        Args:
             method (string): Optimization method
             maxiter (int, optional): Maximum number of iterations to perform,
                default is 1000
             disp (bool, optional): Set to True to print convergence messages.
        """
        options = dict(disp=disp, maxiter=maxiter)

        result = minimize(
            fun=self._loss_and_grad,
            x0=self._get_param_array(),
            method=method,
            jac=True,
            tol=tol,
            callback=callback,
            options=options,
        )
        return result

    def _predict(self, input_new: TensorType, diag=True):
        """
        Predict hte latent output function at input_new.

        If diag=True, then the outputs are the predictive mean and variance, 
        both of shape [n x dy].
        If diag=False, then we predict the full covariance [n x n]; the mean is 
        the same.

        :return: (TensorType, TensorType)
        """
        # diag: is a flag indicates whether only returns the diagonal of
        # predictive variance at input_new
        # :param input_new: np.ndarray
        raise NotImplementedError()

    @input_as_tensor
    def predict_f(self, input_new, diag=True, **kwargs):
        """
        Computes the mean and variance of the latent function at input_new
        return the diagonal of the cov matrix
        Args:
            input_new (numpy.ndarray or TensorType)
        """
        return self._predict(input_new, diag=diag, **kwargs)

    @input_as_tensor
    def predict_y(self, input_new, diag=True, **kwargs):
        """
        Computes the mean and variance of observations at new inputs
        Args:
            input_new (numpy.ndarray)
        """
        mean_f, cov_f = self._predict(input_new, diag=diag, **kwargs)

        if diag:
            return self.likelihood.predict_mean_variance(mean_f, cov_f)
        else:
            return self.likelihood.predict_mean_covariance(mean_f, cov_f)

    @input_as_tensor
    def predict_f_samples(self, input_new, n_samples=1, **kwargs):
        """
        Return [n_samp x n_test x d_y] matrix of samples
        :param input_new:
        :param n_samples:
        :return:
        """
        mu, sigma = self.predict_f(input_new, diag=False, **kwargs)
        chol_s = cholesky(sigma)
        samp = mu + chol_s[None, :, :] @ torch.randn(
            n_samples, *mu.shape, dtype=torch_dtype, device=chol_s.device
        )
        return samp

    @input_as_tensor
    def predict_y_samples(self, input_new, n_samples=1, **kwargs):
        """
        Return [n_samp x n_test x d_y] matrix of samples
        :param input_new:
        :param n_samples:
        :return:
        """
        mu, sigma = self.predict_y(input_new, diag=False, **kwargs)
        chol_s = cholesky(sigma)
        samp = mu + chol_s[None, :, :] @ torch.randn(
            n_samples, *mu.shape, dtype=torch_dtype, device=chol_s.device
        )
        return samp

    def cuda(self):
        """
        Since a GP model "is" its data, we need to make sure that the data make
        it to the GPU when we do model.cuda()
        """

        super().cuda()
        self._data_to_cuda()

    def cpu(self):
        """
        Since a GP model "is" its data, we need to make sure that the data make
        it to the CPU when we do model.cpu()
        """

        super().cpu()
        self._data_to_cpu()

    def _data_to_cpu(self):
        self.X = self.X.cpu()
        self.Y = self.Y.cpu()

    def _data_to_cuda(self):
        self.X = self.X.cuda() 
        self.Y = self.Y.cuda()

    def _loss(self, *args, **kwargs):
        return -(self.log_likelihood(*args, **kwargs) + self.log_prior())



# File: test_sparse_gpr.py
# File Created: Sunday, 1st September 2019 9:14:58 am
# Author: Steven Atkinson (steven@atkinson.mn)

import os
import sys

import pytest
import numpy as np
import torch

from gptorch.models.sparse_gpr import VFE, SVGP
from gptorch.kernels import Matern32
from gptorch import likelihoods
from gptorch import mean_functions
from gptorch.util import torch_dtype, TensorType

from .common import gaussian_predictions

_data_dir = os.path.join(os.path.dirname(__file__), "..", "data", "models", 
    "sparse_gpr")

base_path = os.path.join(os.path.dirname(__file__), "..", "..")
if not base_path in sys.path:
    sys.path.append(base_path)

from test.util import needs_cuda


def atleast_col(func):
    """
    Decorator making sure that matrices loaded are column vectors if 1D
    """

    def wrapped():
        outputs = func()
        if isinstance(outputs, tuple):
            outputs = [o[:, np.newaxis] if o.ndim == 1 else o for o in outputs]
        else:
            outputs = outputs[:, np.newaxis] if outputs.ndim == 1 else outputs
        return outputs

    return wrapped


def _get_matrix(name):
    return np.loadtxt(os.path.join(_data_dir, name + ".dat"))


class _InducingData(object):
    """
    A few pieces in common with these models
    """

    @staticmethod
    @atleast_col
    def _xy():
        return _get_matrix("x"), _get_matrix("y")
    
    @staticmethod
    @atleast_col
    def _x_test():
        return _get_matrix("x_test")
    
    @staticmethod
    @atleast_col
    def _z():
        return _get_matrix("z")


class TestVFE(_InducingData):
    def test_init(self):
        x, y = _InducingData._xy()
        kernel = Matern32(x.shape[1], ARD=True)
        VFE(x, y, kernel)
        VFE(x, y, kernel, inducing_points=_InducingData._z())

        # TODO mean

    def test_compute_loss(self):
        x, y = _InducingData._xy()
        z = _InducingData._z()
        kernel = Matern32(1)
        kernel.length_scales.data = torch.zeros(1, dtype=torch_dtype)
        kernel.variance.data = torch.zeros(1, dtype=torch_dtype)
        likelihood = likelihoods.Gaussian(variance=1.0)

        model = VFE(x, y, kernel, inducing_points=z, likelihood=likelihood,
            mean_function=mean_functions.Zero(1))
        loss = model.loss()
        assert isinstance(loss, torch.Tensor)
        assert loss.ndimension() == 0
        # Computed while I trust the result.
        assert loss.item() == pytest.approx(8.842242323920674)

        # Test ability to specify x and y
        loss_xy = model.loss(x=TensorType(x), y=TensorType(y))
        assert isinstance(loss_xy, torch.Tensor)
        assert loss_xy.item() == loss.item()

        with pytest.raises(ValueError):
            # Size mismatch
            model.loss(x=TensorType(x[:x.shape[0] // 2]))

    @needs_cuda
    def test_compute_loss_cuda(self):
        model = self._get_model()
        model.cuda()
        loss = model.loss()
        assert loss.is_cuda

    def test_predict(self):
        """
        Just the ._predict() method
        """

        x, y = _InducingData._xy()
        z = _InducingData._z()
        kernel = Matern32(1)
        kernel.length_scales.data = torch.zeros(1, dtype=torch_dtype)
        kernel.variance.data = torch.zeros(1, dtype=torch_dtype)
        likelihood = likelihoods.Gaussian(variance=1.0)

        model = VFE(x, y, kernel, inducing_points=z, likelihood=likelihood,
            mean_function=mean_functions.Zero(1))

        x_test = torch.Tensor(_InducingData._x_test())
        mu, s = TestVFE._y_pred()
        gaussian_predictions(model, x_test, mu, s)

    @needs_cuda
    def test_predict_cuda(self):
        model = self._get_model()
        model.cuda()

        x_test = torch.randn(4, model.input_dimension, dtype=torch_dtype).cuda()
        for t in model._predict(x_test):
            assert t.is_cuda

    @staticmethod
    @atleast_col
    def _y_pred():
        return _get_matrix("vfe_y_mean"),  _get_matrix("vfe_y_cov")  

    @staticmethod
    def _get_model():
        x, y = _InducingData._xy()
        z = _InducingData._z()
        kernel = Matern32(1)
        kernel.length_scales.data = torch.zeros(1, dtype=torch_dtype)
        kernel.variance.data = torch.zeros(1, dtype=torch_dtype)
        likelihood = likelihoods.Gaussian(variance=1.0)

        model = VFE(x, y, kernel, inducing_points=z, likelihood=likelihood,
            mean_function=mean_functions.Zero(1))

        return model


class TestSVGP(_InducingData):
    def test_init(self):
        x, y = _InducingData._xy()
        kernel = Matern32(x.shape[1], ARD=True)
        SVGP(x, y, kernel)
        SVGP(x, y, kernel, inducing_points=_InducingData._z())

        SVGP(x, y, kernel, mean_function=mean_functions.Constant(y.shape[1]))
        SVGP(x, y, kernel, mean_function=torch.nn.Linear(x.shape[1], y.shape[1]))

    def test_compute_loss(self):
        x, y = _InducingData._xy()
        z = _InducingData._z()
        u_mu, u_l_s = TestSVGP._induced_outputs()
        kernel = Matern32(1)
        kernel.length_scales.data = torch.zeros(1, dtype=torch_dtype)
        kernel.variance.data = torch.zeros(1, dtype=torch_dtype)
        likelihood = likelihoods.Gaussian(variance=1.0)

        model = SVGP(x, y, kernel, inducing_points=z, likelihood=likelihood,
            mean_function=mean_functions.Zero(1))
        model.induced_output_mean.data = TensorType(u_mu)
        model.induced_output_chol_cov.data = model.induced_output_chol_cov.\
            _transform.inv(TensorType(u_l_s))

        loss = model.loss()
        assert isinstance(loss, torch.Tensor)
        assert loss.ndimension() == 0
        # Computed while I trust the result.
        assert loss.item() == pytest.approx(9.534628739243518)

        # Test ability to specify x and y
        loss_xy = model.loss(x=TensorType(x), y=TensorType(y))
        assert isinstance(loss_xy, torch.Tensor)
        assert loss_xy.item() == loss.item()

        with pytest.raises(ValueError):
            # Size mismatch
            model.loss(x=TensorType(x[:x.shape[0] // 2]), y=TensorType(y))

        model_minibatch = SVGP(x, y, kernel, batch_size=1)
        loss_mb = model_minibatch.loss()
        assert isinstance(loss_mb, torch.Tensor)
        assert loss_mb.ndimension() == 0

        model_full_mb = SVGP(x, y, kernel, inducing_points=z, likelihood=likelihood,
            mean_function=mean_functions.Zero(1), batch_size=x.shape[0])
        model_full_mb.induced_output_mean.data = TensorType(u_mu)
        model_full_mb.induced_output_chol_cov.data = model_full_mb.induced_output_chol_cov.\
            _transform.inv(TensorType(u_l_s))
        loss_full_mb = model_full_mb.loss()
        assert isinstance(loss_full_mb, torch.Tensor)
        assert loss_full_mb.ndimension() == 0
        assert loss_full_mb.item() == pytest.approx(loss.item())

        model.loss(model.X, model.Y)  # Just make sure it works!

    @needs_cuda
    def test_compute_loss_cuda(self):
        model = self._get_model()
        model.cuda()

        loss = model.loss()
        assert loss.is_cuda

    def test_predict(self):
        """
        Just the ._predict() method
        """

        x, y = _InducingData._xy()
        z = _InducingData._z()
        u_mu, u_l_s = TestSVGP._induced_outputs()
        kernel = Matern32(1)
        kernel.length_scales.data = torch.zeros(1, dtype=torch_dtype)
        kernel.variance.data = torch.zeros(1, dtype=torch_dtype)
        likelihood = likelihoods.Gaussian(variance=1.0)

        model = SVGP(x, y, kernel, inducing_points=z, likelihood=likelihood,
            mean_function=mean_functions.Zero(1))
        model.induced_output_mean.data = TensorType(u_mu)
        model.induced_output_chol_cov.data = model.induced_output_chol_cov.\
            _transform.inv(TensorType(u_l_s))

        x_test = TensorType(_InducingData._x_test())
        mu, s = TestSVGP._y_pred()
        gaussian_predictions(model, x_test, mu, s)

    @needs_cuda
    def test_predict_cuda(self):
        model = self._get_model()
        model.cuda()

        x_test = torch.randn(4, model.input_dimension, dtype=torch_dtype).cuda()
        for t in model._predict(x_test):
            assert t.is_cuda

    @staticmethod
    @atleast_col
    def _induced_outputs():
        return _get_matrix("q_mu"), _get_matrix("l_s")

    @staticmethod
    @atleast_col
    def _y_pred():
        return _get_matrix("svgp_y_mean"), _get_matrix("svgp_y_cov")

    @staticmethod
    def _get_model():
        x, y = _InducingData._xy()
        z = _InducingData._z()
        u_mu, u_l_s = TestSVGP._induced_outputs()
        kernel = Matern32(1)
        kernel.length_scales.data = torch.zeros(1, dtype=torch_dtype)
        kernel.variance.data = torch.zeros(1, dtype=torch_dtype)
        likelihood = likelihoods.Gaussian(variance=1.0)

        model = SVGP(x, y, kernel, inducing_points=z, likelihood=likelihood,
            mean_function=mean_functions.Zero(1))
        model.induced_output_mean.data = TensorType(u_mu)
        model.induced_output_chol_cov.data = model.induced_output_chol_cov.\
            _transform.inv(TensorType(u_l_s))
        
        return model



#
# May 10, 2017  Yinhao Zhu
# Utilities: data loading, plotting, batching...
from torch.autograd import Variable
import torch
import numpy as np
from warnings import warn
from scipy.cluster.vq import kmeans2


TensorType = torch.DoubleTensor
torch_dtype = torch.double


def as_tensor(x):
    """
    Convert a numpy matrix into a Tensor.

    Args:
        x (np.ndarray)
    Returns:
        (TensorType)
    """
    if isinstance(x, torch.Tensor):
        return x.type(TensorType)
    if isinstance(x, np.ndarray):
        return TensorType(x)
    elif isinstance(x, float):
        return TensorType([x])
    else:
        raise TypeError("Unsupported type {}".format(type(x)))


def kmeans_centers(x: np.ndarray, k: int, perturb_if_fail: bool=False) -> \
        np.ndarray:
    """
    Use k-means clustering and find the centers of the clusters.
    :param x: The data
    :param k: Number of clusters
    :param perturb_if_fail: Move the points randomly in case of a numpy 
        LinAlgError.
    :return: the centers
    """
    try:
        return kmeans2(x, k)[0]
    except np.linalg.LinAlgError:
        x_scale = x.std(axis=0)
        x_perturbed = x + 1.0e-4 * x_scale * np.random.randn(*x.shape)
        return kmeans2(x_perturbed, k)[0]


def PCA(X, q):
    """
    A helpful function for linearly reducing the dimensionality of the data X
    to q.

    Args:
        X (np.ndarray(n, p)): Observations
        q (int): Dimensionality of latent variable

    Returns:
        (np.ndarray(n, q)): PCA projection array of Z

    """
    assert q <= X.shape[1], "Cannot have more latent dimensions than observed"
    evals, evecs = np.linalg.eigh(np.cov(X.T))
    indices = np.argsort(evals)[::-1]
    W = evecs[:, indices]
    W = W[:, :q]
    return (X - X.mean(0)).dot(W)


def squared_distance(x1: TensorType, x2: TensorType = None) -> TensorType:
    """
    Given points x1 [n1 x d1] and x2 [n2 x d2], return a [n1 x n2] matrix with
    the pairwise squared distances between the points.

    Entry (i, j) is sum_{j=1}^d (x_1[i, j] - x_2[i, j]) ^ 2
    """
    if x2 is None:
        return squared_distance(x1, x1)
    x1s = x1.pow(2).sum(1, keepdim=True)
    x2s = x2.pow(2).sum(1, keepdim=True)
    r2 = x1s + x2s.t() -2.0 * x1 @ x2.t()
    # Prevent negative squared distances using torch.clamp
    # NOTE: Clamping is for numerics.
    # This use of .detach() is to avoid breaking the gradient flow.
    return r2 - (torch.clamp(r2, max=0.0)).detach()



# File: param.py
# File Created: Wednesday, 10th July 2019 11:41:35 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
param.py: Parameters
"""

import torch
from torch.distributions.transforms import ComposeTransform


class Param(torch.nn.Parameter):
    """
    Customized Parameter class extending the PyTorch Parameter class.
    Its main purpose is to include the following additional functionality:
    1) The .transform() member function, in order to impose constraints on the 
        parameter.  Use torch.distributions.transforms classes for this.
    2) the .prior member, for incorporation into joint log-probabilities (e.g. 
        for training)
    """

    def __new__(cls, data=None, requires_grad=True, transform=None, prior=None):
        transform = Param._validate_transform(transform)
        data = transform.inv(data)
        return super().__new__(cls, data, requires_grad=requires_grad)

    def __init__(self, data, requires_grad=True, transform=None, prior=None):
        super().__init__()
        transform = Param._validate_transform(transform)
        self._transform = transform
        self.prior = prior

    def transform(self):
        return self._transform(self)

    def __repr__(self):
        return "Parameter containing:" + self.data.__repr__()

    @staticmethod
    def _validate_transform(t):
        """
        Ensure that the provided transform can be evaluated.

        :param t: The transform to be validated
        
        :return: (torch.distributions.Transform) a valid transform.
        """

        return ComposeTransform([]) if t is None else t



# File: test_util.py
# File Created: Wednesday, 23rd October 2019 2:54:13 pm
# Author: Steven Atkinson (212726320@ge.com)

import os
import sys

import pytest
import torch

base_path = os.path.join(os.path.dirname(__file__), "..")
if not base_path in sys.path:
    sys.path.append(base_path)

from gptorch import util


class TestSquaredDistance(object):
    """
    util.squared_distance

    There's a lot riding on getting this right!
    """

    def test_type(self):
        x1, x2 = self._vals_1d()

        r2_actual = util.squared_distance(x1, x2)
        assert isinstance(r2_actual, util.TensorType)

    def test_shape(self):
        x1, x2 = self._vals_1d()

        r2_actual = util.squared_distance(x1, x2)
        assert r2_actual.shape[0] == x1.shape[0]
        assert r2_actual.shape[1] == x2.shape[0]

    def test_values(self):
        x1, x2 = self._vals_1d()

        r2_actual = util.squared_distance(x1, x2)
        r2_expected = util.TensorType([[0.0, 4.0, 16.0], [1.0, 1.0, 9.0], [4.0, 0.0, 4.0]])
        assert all([a.item() == pytest.approx(e.item()) 
            for a, e in zip(r2_actual.flatten(), r2_expected.flatten())])

    def test_grads_1_nonzero(self):
        """
        Ensure correct first derivatives.
        No major challenges here...
        Case where it's nonzero (point1 and point2 are distinct)
        """
        x1, x2 = self._vals_1d()
        x1.requires_grad_(True)

        r2_actual = util.squared_distance(x1, x2)

        # val = (0-2)^2, grad=2(0-2)=-4
        r2_actual[0, 1].backward(retain_graph=True)  
        grad01_actual = x1.grad[0].item()
        grad01_expected = -4.0
        assert grad01_actual == grad01_expected

    def test_grads_1_zero(self):
        """
        Ensure correct first derivatives.
        No major challenges here...
        Case where it's zero (point1 = point2)
        """
        x1, x2 = self._vals_1d()
        x1.requires_grad_(True)

        r2_actual = util.squared_distance(x1, x2)

        # val = (0-0)^2, grad=2(0-0)=0
        r2_actual[0, 0].backward(retain_graph=True)
        grad00_actual = x1.grad[0].item()
        grad00_expected = 0.0
        assert grad00_actual == grad00_expected

    def test_grads_2(self):
        """
        Ensure correct second derivatives.
        This is nontrivial because the gradient at squared distance~0 could get 
        erased by a .clamp() if you code it up wrong!
        """

        x1, x2 = self._vals_1d()
        x1.requires_grad_(True)
         # Trick to track grads to single rows:
        x1_rows = [xi for xi in x1]
        x1 = torch.stack(x1_rows)

        r2_actual = util.squared_distance(x1, x2)
        r2_00 = r2_actual[0, 0]
        # First derivative:
        # Returns a tuple w/ one entry that's a tensor of size (1,)
        drdx = torch.autograd.grad(r2_00, x1_rows[0], create_graph=True)
        # Second derivative:
        d2rdx2_actual = torch.autograd.grad(drdx[0][0], x1_rows[0])

        # val = (x1-x2)^2
        # 1st der = 2*(x1-x2) = 2(0-0) = 0
        # 2nd der = 2  <-- but this will show up as zero instead if it got clamped!
        d2rdx2_expected = 2.0

        assert d2rdx2_actual[0].item() == d2rdx2_expected

    @staticmethod
    def _vals_1d():
        x1 = util.TensorType([[0.0], [1.0], [2.0]]) + 1.0 / 65.0
        x2 = util.TensorType([[0.0], [2.0], [4.0]]) + 1.0 / 65.0

        return x1, x2



# Yinhao Zhu
# yzhu10@nd.edu
# April 16, 2017

"""
Basic model and parameter classes for Gaussian Processes, inheriting from
:class:`torch.nn.Module` and :class:`torch.nn.Parameter` respectively.
"""

from torch.autograd import gradcheck
import torch
import numpy as np
from time import time
from warnings import warn

from .functions import cholesky
from .param import Param
from .util import TensorType


def _addindent(s_, numSpaces):
    s = s_.split("\n")
    # dont do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * " ") + line for line in s]
    s = "\n".join(s)
    s = first + "\n" + s
    return s


class Model(torch.nn.Module):
    """
    Customized Model class for all GP objects
    """

    def forward(self):
        return None

    def __repr__(self):
        tmpstr = self.__class__.__name__ + " (\n"
        for name, param in self._parameters.items():
            tmpstr = tmpstr + name + "\n" + str(param.transform().data) + "\n"
        for key, module in self._modules.items():
            modstr = module.__repr__()
            modstr = _addindent(modstr, 2)
            tmpstr = tmpstr + "  (" + key + "): " + modstr + "\n"
        tmpstr = tmpstr + ")" + "\n"
        return tmpstr

    # Three functions wrap the scipy.optimize.minimize
    # get_param_array
    # set_parameters
    # loss_and_grad
    def _get_param_array(self):
        """Returns a 1D array by flattening and concatenating all the parameters
        in the model.
        """
        param_array = []
        for param in self.parameters():
            if param.requires_grad:
                param_array.append(param.detach().cpu().numpy().flatten())

        return np.concatenate(param_array)

    def _set_parameters(self, param_array):
        """Set the parameters from a parameter array in the format of
        the return of get_param_array().
        """
        # assert isinstance(param_array, np.ndarray)
        idx_current = 0
        for param in self.parameters():
            if param.requires_grad:
                idx_next = idx_current + param.numel()
                param_in = TensorType(
                    np.reshape(param_array[idx_current:idx_next], param.shape)
                )
                if param.is_cuda:
                    param_in = param_in.cuda()
                param.data = param_in
                idx_current = idx_next

    def _loss_and_grad(self, param_array):
        """
        Computes the loss and gradients for the parameters with values
        ```param_array```.

        1) Take param_array and expand it into the model
        2) Compute the loss and grad
        3) Put the results into numpy data structures

        Note: Any grad entries that are infinite are replaced with zeroes.

        Args:
            param_array (np.ndarray): the state x

        Returns:
            f(x) (np.float64), the objective function values
            g(x) (np.ndarray of np.float64's), the gradient
        """

        # 1) set parameters of the model from a 1D param_array
        self._set_parameters(param_array)
        # print(self)
        for name, param in self.named_parameters():
            # param.grad will be None if either (1) this is the first time we've
            # visited or (2) this param doesn't require a gradient
            if param.grad is not None:
                param.grad.data.zero_()

        # 2) Compute
        loss = self.loss()
        loss.backward()

        # 3) Return as the proper numpy types
        grad = []
        for name, param in self.named_parameters():
            if param.requires_grad:
                grad.append(param.grad.cpu().numpy().flatten())
        print("loss: %s" % loss.item())
        grad = np.concatenate(grad)
        grad_isfinite = np.isfinite(grad)
        ## scipy.optimizer.minimize.L-BFGS-B expects double precision (Fortran)
        if np.all(grad_isfinite):
            return float(loss.item()), grad.astype(np.float64)
        else:
            # self._previous_x = x  # store the last known good value
            print("Warning: inf or nan in gradient: replacing with zeros")
            return (
                loss.item(),
                np.where(grad_isfinite, grad, 0.0).astype(np.float64),
            )

    # Functions for using gradcheck on your model.
    # Use self.loss as the function to be grad-checked, and provide the inputs
    # with self.extract_params
    def extract_params(self):
        """
        Returns:
            (Tuple of TensorTypes)
        """
        return tuple([x for x in self.parameters()])

    def expand_params(self, *args):
        """
        Args:
            args (tuple of TensorTypes): (Get from self.extract_params())
        """
        for arg, (param_name, param) in zip(args, self.named_parameters()):
            if isinstance(arg, Param):
                param.data = arg.data
            elif isinstance(arg, np.ndarray):
                raise NotImplementedError(
                    "Unresolved issues with expanding numpy arrays"
                )

    def log_prior(self):
        """
        Compute the log prior of the model

        Searches through all of the model's parameters and evaluates the 
        log-probability on everything that has a prior.

        :return: (TensorType) The log-prior
        """

        log_prior = 0.0
        for param in self.parameters():
            if hasattr(param, "prior") and param.prior is not None:
                if hasattr(param, "transform") and param.transform is not None:
                    val = param.transform()
                else:
                    val = param.data
                log_prior += param.prior.log_prob(val).sum()

        return log_prior

    def loss(self, *loss_args, params=None, **loss_kwargs):
        """
        Loss, given args to be expanded into the model.

        :param params: The state of the module.  See `.extract_params()`. If 
        provided, the model state is updated before evaluating the loss.
        :type params: tuple of gptorch.util.TensorType objects
        :param loss_args: arguments to pass to the child class's `._loss()` 
        method.
        :type loss_args: tuple
        :param loss_kwargs: Keyword arguments to pass to the child class's 
        `._loss()` method.
        :type loss_kwargs: dict
        """

        if params is not None:
            self.expand_params(*params)
        
        return self._loss(*loss_args, **loss_kwargs)

    def gradcheck(self, eps=1e-6, atol=1e-5, rtol=1e-3, verbose=False):
        """
        This performs a gradcheck on your model's self.compute_loss() function

        See torch.autograd.gradcheck for more on eps, atol, and rtol.

        Args:
            eps (float): Finite difference step size
            atol (float): absolute tolerance (takes over with grads near zero)
            rtol (float): relative tolerance (takes over with large grads)
            verbose (bool): Output more
        Returns:
             (bool) whether the gradcheck passed (true = good)
        """
        if verbose:
            warn("Verbose not yet figured out")
        return gradcheck(
            self.loss, self.extract_params(), eps=eps, atol=atol, rtol=rtol
        )

    def _loss(self, *args, **kwargs):
        raise NotImplementedError("Implement loss function")



# File: test_param.py
# File Created: Wednesday, 10th July 2019 11:47:40 pm
# Author: Steven Atkinson (steven@atkinson.mn)

import pytest
import numpy as np
import torch
from torch.distributions import transforms

from gptorch.param import Param


class TestParam(object):
    def test_init(self):
        x = torch.eye(3) + torch.ones(3, 3)

        Param(x)
        Param(x, transform=transforms.ExpTransform())
        Param(x, transform=transforms.LowerCholeskyTransform())

    def test_access(self):
        """
        Test accessing the value.
        """
        p = Param(torch.DoubleTensor([1.0]))
        assert isinstance(p.data, torch.DoubleTensor)
        assert isinstance(p.data.numpy(), np.ndarray)

    def test_transform_inverse(self):
        x = torch.rand(3, 3)
        t = transforms.ExpTransform()
        p = Param(x, transform=t)

        expected_data = t.inv(x)
        actual_data = p.data
        assert all(
            [
                e.data.numpy() == pytest.approx(a.data.numpy())
                for e, a in zip(expected_data.flatten(), actual_data.flatten())
            ]
        )

    def test_transform_forward(self):
        x = torch.rand(3, 3)
        t = transforms.ExpTransform()
        p = Param(x, transform=t)

        actual_forward = p.transform()
        assert all(
            [
                e.data.numpy() == pytest.approx(a.data.numpy())
                for e, a in zip(x.flatten(), actual_forward.flatten())
            ]
        )



# File: __init__.py
# Author: Yinhao Zhu (yzhu10@nd.edu)

from .base import GPModel

"""
Core Gaussian process models.

Models typically comprise of a GP-based (latent) function combined with some 
likelihood whose density can be evaluated at target (output) locations.

GPR implements (exact) Guassian process regression

sparse_gpr contains the variational free energy (VFE) model of Titsias (2009)
and the sparse variational GP (SVGP) model based on Hensman et al. (2015).
The latter is special in that it is tractably compatible with non-conjugate 
likelihoods, making it usable for non-regression tasks (e.g. classification).
"""

from .gpr import GPR
from .sparse_gpr import VFE, SVGP



# File: test_model.py
# File Created: Saturday, 23rd February 2019 8:30:36 pm
# Author: Steven Atkinson (steven@atkinson.mn)

"""
Tests for model.py
"""

import os
import sys
from functools import partial

import numpy as np
import torch
from torch.distributions import Normal

base_path = os.path.join(os.path.dirname(__file__), "..")
if not base_path in sys.path:
    sys.path.append(base_path)

from gptorch.model import Model
from gptorch.param import Param
from gptorch.util import TensorType, torch_dtype

zeros = partial(torch.zeros, dtype=torch_dtype)
ones = partial(torch.ones, dtype=torch_dtype)


def _standard_normal(*shape):
    return Normal(zeros(*shape), ones(*shape))


class _MockModel(Model):
    """
    Mockup model that has a log-likelihood implemented
    """

    def __init__(self):
        super().__init__()
        self.z = Param(zeros(1), prior=_standard_normal(1))
        self.targets = zeros(1)  # Mock training data

    def log_likelihood(self):
        return _standard_normal(1).log_prob(self.targets).sum()

    def _loss(self):
        return -(self.log_likelihood() + self.log_prior())


class TestModel(object):
    """
    Tests for the Model class
    """

    def test_get_param_array(self):

        model = _MockModel()
        param_array = model._get_param_array()

    def test_set_param_array(self):

        model = _MockModel()
        param_array = model._get_param_array()
        param_array += 1.0
        model._set_parameters(param_array)

    def test_loss_and_grad(self):
        model = _MockModel()
        param_array = model._get_param_array()
        loss, grad = model._loss_and_grad(param_array)

    def test_extract_params(self):

        model = _MockModel()
        params = model.extract_params()

    def test_expand_params(self):

        model = _MockModel()
        params = model.extract_params()
        model.expand_params(params)

    def test_log_likelihood(self):
        model = _MockModel()

        target = zeros(1)
        log_likelihood = model.log_likelihood()

        assert isinstance(log_likelihood, TensorType)
        assert log_likelihood.ndimension() == 0
        # See _MockModel to verify target is 0.0, likelihood is a standard normal.
        assert log_likelihood.item() == _standard_normal(1).log_prob(zeros(1)).sum().item()

    def test_log_prior(self):
        """
        Basic test to make sure that log_prior() works in a basic model: 
        Only one parameter, it's 1D, and it has a prior.
        """

        model = _MockModel()
        
        log_prior = model.log_prior()
        assert isinstance(log_prior, TensorType)
        assert log_prior.ndimension() == 0
        # See _MockModel:
        assert log_prior.item() == _standard_normal(1).log_prob(zeros(1)).sum().item()

    def test_loss(self):
        model = _MockModel()

        loss = model.loss()

        assert isinstance(loss, TensorType)
        assert loss.ndimension() == 0
        assert loss == -(model.log_likelihood() + model.log_prior())



#
# Apr 15, 2017  Yinhao Zhu
#
# Class for GP regression

"""
Class for Vanilla GP regression.
"""

import torch
import numpy as np

from .. import kernels
from ..model import Param
from .. import likelihoods
from ..functions import cholesky, inverse, lt_log_determinant, trtrs
from ..util import TensorType
from .base import GPModel


class GPR(GPModel):
    """
    Gaussian Process Regression
    """

    def __init__(self, x, y, kernel, mean_function=None, likelihood=None, name="gpr"):
        """
        Default likelihood is Gaussain, mean function is zero.

        Args:
            observations (np.ndarray or TensorType): a matrix of observed outputs.
                Each datum is a row in the matrix (# rows = # training data,
                # columns = output dimensionality)
            input (np.ndarray or TensorType): the observed inputs.
                If input is a numpy matrix or torch TensorType, then each datum is
                a row in the matrix (# rows = # training data,
                # columns = input dimensionality).
            kernel (Kernel): The kernel function for computing the covariance
                matrices
            mean_function (MeanFunction): class for populating the design matrix
                for computing the parameterized mean function.
            likelihood (Likelihood): A likelihood model
        """
        
        super().__init__(x, y, kernel, likelihood, mean_function, name)

    def log_likelihood(self, x=None, y=None):
        """
        Loss is equal to the negative of the prior log likelihood

        Adapted from Rasmussen & Williams, GPML (2006), p. 19, Algorithm 2.1.
        """

        x = x if x is not None else self.X
        y = y if y is not None else self.Y
        if not x.shape[0] == y.shape[0]:
            raise ValueError("X and Y must have same # data.")

        num_input, dim_output = y.shape

        L = cholesky(self._compute_kyy(x=x))
        alpha = trtrs(y - self.mean_function(x), L)
        const = TensorType([-0.5 * dim_output * num_input * np.log(2 * np.pi)])
        if alpha.is_cuda:
            const = const.cuda()  # TODO cache this?
        loglik = -0.5 * alpha.pow(2).sum() - dim_output * lt_log_determinant(L) + const
        return loglik

    def _compute_kyy(self, x=None):
        """
        Computes the covariance matrix over the training inputs

        Returns:
            Ky (TensorType)
        """

        x = x if x is not None else self.X
        num_input = x.shape[0]

        return (
            self.kernel.K(x)
            + (self.likelihood.variance.transform())
            .expand(num_input, num_input)
            .diag()
            .diag()
        )

    def _predict(self, x_new: TensorType, diag=True, x=None):
        """
        This method computes

        .. math::
            p(F^* | Y )

        where F* are points on the GP at x_new, Y are observations at the
        input X of the training data.
        :param x_new: test inputs; should be two-dimensional
        """

        x = x if x is not None else self.X

        k_ys = self.kernel.K(x, x_new)

        L = cholesky(self._compute_kyy(x=x))
        A = trtrs(k_ys, L)
        V = trtrs(self.Y - self.mean_function(x), L)
        mean_f = A.t() @ V + self.mean_function(x_new)

        if diag:
            var_f = (
                self.kernel.Kdiag(x_new) - 
                (A * A).sum(0)
            )[:, None].expand_as(mean_f)
        else:
            var_f = self.kernel.K(x_new) - A.t() @ A

        return mean_f, var_f




"""
Likelihood classes

Objects that propagate either points of distributions through a transformation
yielding a (likelihood) distribution to be evaluated at observed targets 
(outputs) to guide Bayesian inference.

We also desire to be able to compute expectations of log-densities marginalizing 
over input distributions (i.e. "marginal log-likelihoods") as typically occur in
variational inference.
"""

import abc
from math import pi

from torch.nn import Parameter
import torch
from torch import distributions

from .model import Model, Param
from .settings import DefaultPositiveTransform
from .util import torch_dtype

from .util import TensorType


class Likelihood(Model):
    """
    Probabilities that conventionally factorize across data.
    Typically used as the "second stage" of a GP model that goes 
    x -(GP)-> f -(likelihood)-> y
    x = inputs
    f = latent outputs
    y = observed outputs

    We typically have two uses for these:
    1) Do the marginalization in p(y|x) = \int p(y|f) p(f|x) df 
        (e.g. predictions)
    2) Do the marginalization in logp(y) >= \int logp(y|f) p(f) df
        (e.g. variational inference)
    """

    def __init__(self):
        super(Likelihood, self).__init__()

    def predict_mean_variance(self, mean_f, var_f):
        """
        The likelihood of observations y given latent functions f is p(y | f)
        It computes the mean and variance of observations Y given latent
        functions F, which itself is Gaussian.

        p(y) = \int p(y | f) p(f) df

        Gauss-Hermite quadrature is used for numerical integration.

        :param mean_f: mean of latent functions
        :type mean_f: TensorType
        :param var_f: variance of the latent functions
        :type var_f: TensorType
        :return: (TensorType, TensorType) mean and variance of the observations
        """
        # TODO: Gauss-Hermite quadrature
        raise NotImplementedError

    def forward(self):
        return None

    @abc.abstractmethod
    def propagate_log(
        self, qf: torch.distributions.Distribution, targets: torch.Tensor
    ):
        """
        Evaluate the marginal log-likelihood at the targets:
        <log p(y|f)>_q(f)
        Used in variational inference.
        """
        raise NotImplementedError("Implement quadrature fallback")


class Gaussian(Likelihood):
    """
    (Spherical) Gaussian likelihood p(y|f)
    """

    def __init__(self, variance=1.0):
        super(Gaussian, self).__init__()
        self.variance = Param(
            TensorType([variance]), transform=DefaultPositiveTransform()
        )

    def logp(self, F, Y):
        """
        Evaluate the log-density of targets at Y, given a Gaussian density 
        centered at F.

        :param F: Center of the density
        :type F: TensorType
        :param Y: Targets where we want to compute the log-pdf
        :type Y: TensorType
        """
        return distributions.Normal(F, torch.sqrt(self.variance.transform())).log_prob(
            Y
        )

    def predict_mean_variance(self, mean_f, var_f):
        """
        Integrate the input (a Gaussian with provided mean & variance) over the
        likelihood density

        :param mean_f: Mean of input Gaussian
        :type mean_f: TensorType
        :param var_f: Variance of input Gaussian
        :type var_f: TensorType

        :return: (TensorType, TensorType) mean & variance
        """
        # TODO: consider mulit-output case
        # stupid syntax - expecting broadcasting in PyTorch
        return mean_f, var_f + self.variance.transform().expand_as(var_f)

    def predict_mean_covariance(self, mean_f, cov_f):
        return mean_f, cov_f + self.variance.transform().expand_as(cov_f).diag().diag()

    def propagate_log(self, qf, targets):
        if not isinstance(qf, torch.distributions.Normal) and not isinstance(
            qf, torch.distributions.MultivariateNormal
        ):
            raise TypeError("Expect Gaussian q(f)")

        mu, s = qf.loc, qf.variance
        n = targets.nelement()
        if not mu.nelement() == n:
            raise ValueError(
                "Targets (%i) and q(f) (%i) have mismatch in size" % (n, mu.nelement())
            )

        sigma_y = self.variance.transform()

        return -0.5 * (
            n * (torch.log(TensorType([2.0 * pi])).to(sigma_y.device) 
            + torch.log(sigma_y))
            + (torch.sum((targets - mu) ** 2) + s.sum()) / sigma_y
        )



# File: test_mean_functions.py
# File Created: Saturday, 13th July 2019 3:51:40 pm
# Author: Steven Atkinson (steven@atkinson.mn)

import pytest
import torch

from gptorch import mean_functions


class TestConstant(object):
    def test_init(self):
        dy = 2
        mean_functions.Constant(dy)
        mean_functions.Constant(dy, val=torch.randn(dy))

        with pytest.raises(ValueError):
            # dim doesn't match val
            mean_functions.Constant(dy, val=torch.Tensor([1.0]))

    def test_forward(self):
        n, dx, dy = 5, 3, 2

        y = mean_functions.Constant(dy)(torch.rand(n, dx))
        assert isinstance(y, torch.Tensor)
        assert all([e == a for e, a in zip(y.flatten(), torch.zeros(n, dy).flatten())])

        val = torch.randn(dy)
        y = mean_functions.Constant(dy, val=val)(torch.rand(n, dx))
        assert isinstance(y, torch.Tensor)
        assert all(
            [e == a for e, a in zip(y.flatten(), (val + torch.zeros(n, dy)).flatten())]
        )


class TestZero(object):
    def test_init(self):
        mean_functions.Zero(2)

    def test_forward(self):
        n, dx, dy = 5, 3, 2
        y = mean_functions.Zero(dy)(torch.rand(n, dx))
        assert isinstance(y, torch.Tensor)
        assert all([e == a for e, a in zip(y.flatten(), torch.zeros(n, dy).flatten())])
